<?php

declare(strict_types=1);

namespace Doctrine\DBAL\Tests\Functional\DataAccess;

class FetchClass
{
    public int $test_int;

    public string $test_string;

    public string $test_datetime;
}
