<?php

declare(strict_types=1);

namespace Doctrine\DBAL\Tests\Types;

use Doctrine\DBAL\Platforms\AbstractPlatform;
use Doctrine\DBAL\Types\SmallFloatType;
use PHPUnit\Framework\MockObject\Stub;
use PHPUnit\Framework\TestCase;

class SmallFloatTest extends TestCase
{
    private AbstractPlatform&Stub $platform;
    private SmallFloatType $type;

    protected function setUp(): void
    {
        $this->platform = self::createStub(AbstractPlatform::class);
        $this->type     = new SmallFloatType();
    }

    public function testFloatConvertsToPHPValue(): void
    {
        self::assertSame(5.5, $this->type->convertToPHPValue('5.5', $this->platform));
    }

    public function testFloatNullConvertsToPHPValue(): void
    {
        self::assertNull($this->type->convertToPHPValue(null, $this->platform));
    }

    public function testFloatConvertToDatabaseValue(): void
    {
        self::assertSame(5.5, $this->type->convertToDatabaseValue(5.5, $this->platform));
    }

    public function testFloatNullConvertToDatabaseValue(): void
    {
        self::assertNull($this->type->convertToDatabaseValue(null, $this->platform));
    }
}
